//
// eda.c
//
// step response EDA
//    9600 baud FTDI interface
//
// Eyal Perry
// 11/04/19
//
// (c) Massachusetts Institute of Technology 2010
// This work may be reproduced, modified, distributed,
// performed, and displayed for any purpose. Copyright is
// retained and must be preserved. The work is provided
// as is; no warranty is provided, and users accept all
// liability.
//

#include <math.h>
#include <avr/io.h>
#include <util/delay.h>
#include "ws2812_config.h"
#include "light_ws2812.h"

#define output(directions,pin) (directions |= pin) // set port direction for output
#define input(directions,pin) (directions &= (~pin)) // set port direction for input
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 8.5 // bit delay for 115200 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay
#define charge_delay_1() _delay_us(1) // charge delay 1
#define charge_delay_2() _delay_us(10) // charge delay 2
#define charge_delay_3() _delay_us(100) // charge delay 3
#define settle_delay() _delay_us(100) // settle delay
#define char_delay() _delay_ms(10) // char delay

#define BUFFER_SIZE 50
#define PIXEL_NUM 21
#define LIGHT_DECAY_STEP 10


#define serial_port PORTA
#define serial_direction DDRA
#define serial_pins PINA
#define serial_pin_out (1 << PA2)

#define led_port PORTA
#define led_direction DDRA
#define led (1 << PA3)

#define charge_port PORTA
#define charge_direction DDRA
#define charge_pin (1 << PA0)

struct cRGB pixels[PIXEL_NUM];

void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {
   //
   // send character in txchar on port pin
   //    assumes line driver (inverts bits)
   //
   // start bit
   //
   clear(*port,pin);
   bit_delay();
   //
   // unrolled loop to write data bits
   //
   if bit_test(txchar,0)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,1)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,2)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,3)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,4)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,5)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,6)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,7)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   //
   // stop bit
   //
   set(*port,pin);
   bit_delay();
   //
   // char delay
   //
   bit_delay();
   }

int main(void) {
   //
   // main
   //
   unsigned int eda;
   unsigned int load_up, load_down;
   float load;
   float load_filter = 0.0;
   float eda_filter = 0.0;


   unsigned int last_loads[BUFFER_SIZE];
   float load_mean = 0;
   float load_diff = 0;

   unsigned int last_edas[BUFFER_SIZE];
   float eda_mean = 0;
   float eda_diff = 0;

   unsigned char last_idx = 0;

   float fear;
   float last_fear = 0;

   //
   // set clock divider to /1
   //
   CLKPR = (1 << CLKPCE);
   CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);
   // initialize LEDs
   output(led_direction, led);
   clear(led_port, led);
   //
   // initialize output pins
   //
   set(serial_port, serial_pin_out);
   output(serial_direction, serial_pin_out);
   // initialize charge pin
   clear(charge_port, charge_pin);
   output(charge_direction, charge_pin);
   // set arrays to zero
   for (unsigned char i = 0; i < BUFFER_SIZE; i++) {
      last_loads[i] = 0;
      last_edas[i] = 0;
    }
    for (unsigned char i = 0; i < PIXEL_NUM; i++) {
      pixels[i].r = 0;
      pixels[i].g = 0;
      pixels[i].b = 0;
    }
    ws2812_setleds(pixels, PIXEL_NUM);
   //
   // init A/D
   //
   ADMUX = (0 << REFS1) | (0 << REFS0) // Vcc ref
   //    | (0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (1 << MUX2) | (1 << MUX1) | (1 << MUX0); // PA7
         | (0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (0 << MUX2) | (0 << MUX1) | (1 << MUX0); // PA1
   ADCSRA = (1 << ADEN) // enable
      | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // prescaler /128
   ADCSRB |= (0 << ADLAR);  // right adjust

   // set timer prescaler to /1024
   // assuming a clock speed of 20Mhz, a 16 bit timer can count until:
   // 1 / (20 / 1024) * 65536 = 3.3sec~
   // good enough for us, since i only use to to measure taps
   // TCCR1B = (1<<CS10) | (1<<CS12);
   // TCNT1 = 0;
   //
   // main loop
   //
   while (1) {


      // Sense Load
      //
      // settle, charge, and wait 1
      //
      settle_delay();
      set(charge_port, charge_pin);
      charge_delay_2();
      //
      // initiate conversion
      //
      ADMUX = (0 << REFS1) | (0 << REFS0) // Vcc ref
            | (0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (0 << MUX2) | (0 << MUX1) | (1 << MUX0); // PA1
      ADCSRA |= (1 << ADSC);
      //
      // wait for completion
      //
      while (ADCSRA & (1 << ADSC))
         ;
      //
      // save result, ADCH*256 + ADCL
      //
      load_up = ADCL;
      load_up |= ADCH << 8;
      //
      // settle, discharge, and wait 1
      //
      settle_delay();
      clear(charge_port, charge_pin);
      charge_delay_2();
      //
      // initiate conversion
      //
      ADCSRA |= (1 << ADSC);
      //
      // wait for completion
      //
      while (ADCSRA & (1 << ADSC))
         ;
      //
      // save result, ADCH*256 + ADCL
      //
      load_down = ADCL;
      load_down |= ADCH << 8;

      // Now sense EDA
      //
      // initiate conversion
      //
      ADMUX = (0 << REFS1) | (0 << REFS0) // Vcc ref
            | (0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (1 << MUX2) | (1 << MUX1) | (1 << MUX0); // PA7
      ADCSRA |= (1 << ADSC);
      //
      // wait for completion
      //
      while (ADCSRA & (1 << ADSC))
         ;
      //
      // save result, ADCH*256 + ADCL
      //
      eda = ADCL;
      eda |= ADCH << 8;


      // calculate load difference
      load = (load_up + (1023 - load_down)) / 2;
      // load smoothing
      load_filter = (0.5*load_filter) + (0.5*load);
      // calculate difference from mean
      load_diff = load_filter - load_mean;
      // update the mean
      load_mean -= last_loads[last_idx] / (float) (BUFFER_SIZE);
      last_loads[last_idx] = load_filter;
      load_mean += last_loads[last_idx] / (float) (BUFFER_SIZE);

      // eda smoothing
      eda_filter = (0.5*eda_filter) + (0.5*eda);
      // calculate difference from mean
      eda_diff = eda_filter - eda_mean;
      // update the mean
      eda_mean -= last_edas[last_idx] / (float) (BUFFER_SIZE);
      last_edas[last_idx] = eda_filter;
      eda_mean += last_edas[last_idx] / (float) (BUFFER_SIZE);

      // increment the counter (or rotate it back)
      if (++last_idx >= BUFFER_SIZE)
        last_idx = 0;

      // if no one is holding the handle, set the timer to zero
      if (eda_mean < 1.0) {
        // TCNT1 = 0;
        clear(led_port, led);
      } else {
        set(led_port, led);
      }

      // the new fear value will set the brightness of the first led, and we push the other leds up the stream
      for (unsigned char i = PIXEL_NUM - 1; i > 0; i--) {
        pixels[i].b = pixels[i - 1].b;
      }
      fear = 0;

      // // if the handle is being held for approx 1sec, get new fear value
      // if (TCNT1 > 64000) {
      //   fear = fmin(144, 4 * (pow(fmax(0.0, eda_diff + fmin(0.0, 1.1 * load_diff)), 2) - 2.0));
      // // }
      fear = fmin(144, pow(fmax(0.0, eda_diff + fmin(0.0, 2 * load_diff) - 6.0), 2));
      if (fear < 16) {
        fear = 0;
      }
      // fear = fmin(255, fmax(0, eda_diff));


      // we add decay, so the new color is the max of new fear value OR the previous value minus decay step
      pixels[0].b = (unsigned char) fmax(fear, last_fear - LIGHT_DECAY_STEP);
      last_fear = pixels[0].b;
      // Light up NeoPixels
      ws2812_setleds(pixels, PIXEL_NUM);
      _delay_ms(30);
    }
 }
